""" Flask-Seeder """

from .seeder import Seeder
from .faker import Faker

# pylint: disable=too-few-public-methods
class SeedConfig:
    """ Seed config stored in Flask

    This is intended to be stored as an instance inside a Flask app
    for access to Flask-Seed configuration.
    """
    def __init__(self, db):
        self.db = db


# pylint: disable=too-few-public-methods
class FlaskSeeder:
    """ Flask extension Flask-Seeder

    Usage:
        seeder = FlaskSeeder()
        seeder.init_app(app, db)

    """
    def __init__(self, app=None, db=None):
        """ Initialize FlaskSeeder
        Arguments:
            app: Flask app
            db: SQLAlchemy database object
        """
        self.app = app
        self.db = db

        if app is not None and db is not None:
            self.init_app(app, db)

    def init_app(self, app, db=None):
        """ Initialize app after construction

        Initialize Flask-Seeder and register as an extension with Flask

        Arguments:
            app: Flask app
            db: SQLAlchemy database object

        """
        self.db = db or self.db

        if not hasattr(app, 'extensions'):
            app.extensions = {}

        app.extensions['flask_seeder'] = SeedConfig(db)



""" String Generator pattern parser """

import re

# pylint: disable=bad-whitespace
TOKEN_MATCHERS = [
    ("CHARCODE",            r"\\([a-z])"),
    ("ONEOF",               r"\[([a-zA-Z0-9]+)\]"),
    ("RANGE",               r"\[(\w)-(\w)\]"),
    ("QUANTIFIER",          r"\{(\d+)\}"),
    ("QUANTIFIER_RANGE",    r"\{(\d*),(\d+)\}"),
    ("STRING_GROUP",        r"\(([^|]([^)]+))\)"),
    ("NUMBER",              r"\d+"),
    ("STRING",              r"[a-zA-Z]+"),
    ("LITERAL",             r"."),
]

def strtype(string):
    """ Check string type

    Arguments:
        string: String to check

    Returns:
        Returns one of the following results
        * DIGIT: String is digit [0-9]
        * ALPHA: String is alpha [a-zA-Z]
        * STRING: If no other match is found, it is assumed to be a STRING type
    """
    if str.isdigit(string):
        return "DIGIT"
    if str.isalpha(string):
        return "ALPHA"
    return "STRING"

class Token: # pylint: disable=too-few-public-methods
    """ Token class
    Used by a tokenizer to store token data

    Attributes:
        group: Token group type or classification
        value: Token value
        pos: Starting position in input stream where token is found
    """

    def __init__(self, group, value, pos=0):
        self.group = group
        self.value = value
        self.pos = pos


class Tokenizer:
    """ String Generator Tokenizer

    Attributes:
        string: String to tokenize
        tokens: List of tokens after successful tokenization
    """

    def __init__(self):
        self.tokens = []
        self.regex = "|".join("(?P<%s>%s)" % matcher for matcher in TOKEN_MATCHERS)
        self._cursor = 0

    def run(self, string=None):
        """ Run tokenizer """
        for match in re.finditer(self.regex, string):
            group = match.lastgroup
            value = match.group()
            self.tokens.append(Token(group, value, match.start()))

    def next(self):
        """ Get next token

        Retrieve next token and move cursor forward.

        Returns:
            A Token instance
        """
        if self._cursor >= len(self.tokens):
            return None

        token = self.tokens[self._cursor]
        self._cursor += 1
        return token

    def peek(self):
        """ Peek at the next token

        Retrieve next token without moving cursor forward.

        Returns:
            A Token instance
        """
        if self._cursor >= len(self.tokens):
            return None

        return self.tokens[self._cursor]

class SGParser:
    """ String Generator Pattern parser """

    def __init__(self, tokenizer=None):
        self.tokenizer = tokenizer

    def _repeat(self):
        ahead = self.tokenizer.peek()
        if ahead and ahead.group == "QUANTIFIER":
            result = self.parse_QUANTIFIER(self.tokenizer.next())
        elif ahead and ahead.group == "QUANTIFIER_RANGE":
            result = self.parse_QUANTIFIER_RANGE(self.tokenizer.next())
        else:
            result = {"type": "QUANTIFIER", "value": 1}

        return result

    def parse_STRING(self, token): # pylint: disable=invalid-name, no-self-use
        """ Parse a STRING token

        Returns:
            Dictionary representing the AST node structure for string
        """
        return {"type": token.group, "value": token.value}

    def parse_RANGE(self, token): # pylint: disable=invalid-name
        """ Parse a RANGE token

        Returns:
            Dictionary representing the AST node structure for range
        """
        (start, end) = re.search(r"\[([a-zA-Z0-9])-([a-zA-Z0-9])\]", token.value).groups()
        if start > end or strtype(start) != strtype(end):
            raise ValueError("Invalid range %s at position %d" % (token.value, token.pos))

        result = {"type": token.group,
                  "value": {
                      "start": start,
                      "end": end,
                  },
                  "repeat": self._repeat()}

        return result

    def parse_QUANTIFIER(self, token): # pylint: disable=invalid-name, no-self-use
        """ Parse a QUANTIFIER token

        Returns:
            Dictionary representing the AST node structure for quantifier
        """
        value = re.match(r"\{(\d+)\}", token.value).group(1)
        return {"type": token.group, "value": int(value)}

    def parse_CHARCODE(self, token): # pylint: disable=invalid-name
        """ Parse a CHARCODE token

        Returns:
            Dictionary representing the AST node structure for charcode
        """
        value = re.match(r"\\([a-z])", token.value).group(1)
        return {"type": token.group, "value": value, "repeat": self._repeat()}

    def parse_ONEOF(self, token): # pylint: disable=invalid-name
        """ Parse a ONEOF token

        Returns:
            Dictionary representing the AST node structure for oneof
        """
        values = re.findall("[a-zA-Z0-9]", token.value)
        return {"type": token.group, "value": values, "repeat": self._repeat()}

    def parse_QUANTIFIER_RANGE(self, token): # pylint: disable=invalid-name, no-self-use
        """ Parse a QUANTIFIER_RANGE token

        Returns:
            Dictionary representing the AST node structure for quantifier range
        """
        (start, end) = re.search(r"\{(\d+),(\d+)\}", token.value).groups()

        if int(start) > int(end):
            raise ValueError("Invalid range %s at %d" % (token.value, token.pos))

        return {"type": token.group, "value": {"start": int(start), "end": int(end)}}

    def parse_STRING_GROUP(self, token): # pylint: disable=invalid-name
        """ Parse a STRING_GROUP token

        Returns:
            Dictionary representing the AST node structure for string group
        """
        options = re.match(r"\((.+)\)", token.value).group(1)
        values = re.split(r"\|", options)
        return {"type": token.group, "value": values, "repeat": self._repeat()}

    def parse_NUMBER(self, token): # pylint: disable=invalid-name
        """ Parse a NUMBER token

        Returns:
            Dictionary representing the AST node structure for number
        """
        return {"type": token.group, "value": int(token.value), "repeat": self._repeat()}

    def parse_LITERAL(self, token): # pylint: disable=invalid-name, no-self-use
        """ Parse a LITERAL token

        Returns:
            Dictionary representing the AST node structure for literal
        """
        return {"type": "LITERAL", "value": token.value}

    def parse(self, string):
        """ Main parser method

        Parses a list of tokens into an AST

        Arguments:
            string: The string to be parsed

        Returns:
            Dictionary representing the String Generator pattern AST
        """
        ast = []
        self.tokenizer.run(string)
        token = self.tokenizer.next()
        while token:
            function_name = "parse_" + token.group
            if not hasattr(self, function_name):
                function_name = "parse_LITERAL"

            func = getattr(self, function_name)
            ast.append(func(token))

            token = self.tokenizer.next()

        return ast



from unittest import TestCase
from unittest.mock import MagicMock, mock_open, patch

from flask_seeder.generator import Name

MOCK_CONTENTS = [
    "name1",
    "name2"
]

class TestNameGenerator(TestCase):

    def setUp(self):
        self.rnd_mock = MagicMock()
        self.generator = Name(rnd=self.rnd_mock)

    @patch("flask_seeder.generator.read_resource", return_value=MOCK_CONTENTS)
    def test_generate_name(self, m_read_resource):
        result = self.generator.generate()

        self.rnd_mock.choice.assert_called_once_with(MOCK_CONTENTS)



from unittest import TestCase
from unittest.mock import MagicMock

from flask_seeder.generator import Integer

class TestIntegerGenerator(TestCase):

    def setUp(self):
        self.rnd_mock = MagicMock()
        self.generator = Integer(rnd=self.rnd_mock)

    def test_generate_int_with_default_values(self):
        self.generator.generate()

        self.rnd_mock.randint.assert_called_once_with(
            self.generator.start,
            self.generator.end
        )

    def test_generate_int_with_custom_values(self):
        self.generator.start = 10
        self.generator.end = 20

        self.generator.generate()
        self.rnd_mock.randint.assert_called_once_with(
            self.generator.start,
            self.generator.end
        )


""" Faker module """

from flask_seeder.generator import Generator

# pylint: disable=too-few-public-methods
class Faker:
    """ Base Faker class

    The `init` attribute is a dictionary that tells Faker how to
    initialize the classes, for example:
        {
            "name": generator.Name()
        }

    Attributes:
        cls: The type of class to be created
        init: Dictionary with initialization data
    """

    def __init__(self, cls=None, init=None):
        """ Initialize faker """
        self.cls = cls
        self.init = init

    def _init_args(self):
        args = {}
        if self.init is None:
            return args

        for arg, value in self.init.items():
            if isinstance(value, Generator):
                args[arg] = value.generate()
            else:
                args[arg] = value

        return args

    def create(self, limit=1):
        """ Create objects

        Create a number of instance of `cls`,
        all initialized with data from `init`.

        Arguments:
            limit: How many objects to create, default 1.

        Returns:
            List of `cls` instances initialized with data from `init`.
        """
        instances = []
        for _ in range(limit):
            args = self._init_args()
            instances.append(self.cls(**args))

        return instances



from unittest import TestCase
from unittest.mock import MagicMock, patch

from flask_seeder.generator import String

class TestStringGenerator(TestCase):

    def setUp(self):
        self.generator = String()

    def test_integer_range_produce_character_from_valid_range(self):
        m_rnd = MagicMock()
        self.generator.rnd = m_rnd
        self.generator.pattern = "[4-6]"

        self.generator.generate()

        m_rnd.choice.assert_called_once_with("456")

    def test_string_range_produce_character_from_valid_range(self):
        m_rnd = MagicMock()
        self.generator.rnd = m_rnd
        self.generator.pattern = "[i-o]"

        self.generator.generate()

        m_rnd.choice.assert_called_once_with("ijklmno")

    def test_integer_range_pattern(self):
        self.generator.pattern = "[0-9]"

        string = self.generator.generate()

        self.assertRegex(string, r"^\d$")

    def test_string_range_pattern(self):
        self.generator.pattern = "[a-z]"

        string = self.generator.generate()

        self.assertRegex(string, "^[a-z]$")

    def test_range_pattern_quantifier(self):
        self.generator.pattern = "[0-9]{5}"

        string = self.generator.generate()

        self.assertRegex(string, r"^\d{5}$")

    def test_multiple_range_patterns(self):
        self.generator.pattern = "[0-9][a-c]"

        string = self.generator.generate()

        self.assertRegex(string, r"^\d[a-c]$")

    def test_multiple_range_patterns_with_quantifiers(self):
        self.generator.pattern = "[0-9]{4}[x-z]{2}"

        string = self.generator.generate()

        self.assertRegex(string, r"^\d{4}[x-z]{2}$")


    def test_integer_character_code(self):
        self.generator.pattern = r"\d"

        string = self.generator.generate()

        self.assertRegex(string, r"^\d$")

    def test_string_character_code(self):
        self.generator.pattern = r"\c"

        string = self.generator.generate()

        self.assertRegex(string, "^[a-zA-Z]$")

    def test_character_code_with_quantifier(self):
        self.generator.pattern = r"\d{3}"

        string = self.generator.generate()

        self.assertRegex(string, r"^\d{3}$")

    def test_oneof_pattern(self):
        self.generator.pattern = r"[aoueiy]"

        string = self.generator.generate()

        self.assertRegex(string, "^[aoueiy]$")

    def test_oneof_pattern_with_quantifier(self):
        self.generator.pattern = r"[aoueiy]{3}"

        string = self.generator.generate()

        self.assertRegex(string, "^[aoueiy]{3}$")

    def test_complex_string_generation(self):
        self.generator.pattern = r"abc\d{4}[i-m]{2}[qwerty]xyz"

        string = self.generator.generate()

        self.assertRegex(string, r"^abc\d{4}[i-m]{2}[qwerty]xyz$")

    def test_charcode_with_quantifier_range(self):
        self.generator.pattern = r"\c{1,2}"

        string = self.generator.generate()

        self.assertRegex(string, "[a-zA-Z]{1,2}")

    def test_invalid_charcode_raise_ValueError(self):
        self.generator.pattern = r"\q"

        with self.assertRaises(ValueError):
            self.generator.generate()

    def test_string_group(self):
        self.generator.pattern = r"(one|two)"

        string = self.generator.generate()

        self.assertRegex(string, r"(one|two)")

    def test_number(self):
        self.generator.pattern = r"123"

        string = self.generator.generate()

        self.assertEqual(string, "123")



""" Base seeder """

# pylint: disable=too-few-public-methods
class Seeder:
    """ Base seeder class """

    def __init__(self, db=None):
        self.db = db
        self.name = None
        self.mod_path = None
        self.file_path = None

    def run(self):
        """ Run the seeder script.

        Must be implemented by the client.
        """
        raise NotImplementedError()



from unittest import TestCase
from unittest.mock import MagicMock, patch, mock_open

from flask_seeder.generator import Generator, resource_path, read_resource, slicer

MOCK_CONTENTS = "line1\nline2"

class TestGenerator(TestCase):

    def setUp(self):
        self.rnd_mock = MagicMock()
        self.generator = Generator(rnd=self.rnd_mock)


    def test_generate_raise_NotImplementedError(self):
        with self.assertRaises(NotImplementedError):
            self.generator.generate()

    @patch("flask_seeder.generator.pkg_resources")
    def test_resource_path(self, m_pkg):
        resource_path("test")

        m_pkg.resource_filename.assert_called_once()

    @patch("flask_seeder.generator.resource_path", return_value="test")
    @patch("flask_seeder.generator.open", mock_open(read_data=MOCK_CONTENTS))
    def test_read_resource_return_contents_as_list(self, m_open):
        expected = [
            "line1",
            "line2"
        ]

        result = read_resource("test")

        self.assertListEqual(result, expected)

    def test_slicer_return_new_string(self):
        original = "abcdefghijklmnopqrstuvwxyz"
        expected = "ijklmno"

        result = slicer(original, "i", "o")

        self.assertEqual(result, expected)

    def test_slicer_return_None_if_no_slice_is_found(self):
        original = "abcdefghijklmnopqrstuvwxyz"

        # Inversed start/end
        result = slicer(original, "z", "a")

        self.assertIsNone(result)



""" Generators module """

import random
import pkg_resources

from flask_seeder.parser import SGParser, Tokenizer

def resource_path(path):
    """ Get the resource path

    Arguments:
        path: Relative path to the resource

    Returns:
        Returns the full filesystem path to the resource.
        Note that no validation is made to ensure the resource actually exist.
    """
    return pkg_resources.resource_filename("flask_seeder", "data/" + path)

def read_resource(path):
    """ Read resource text file

    Reads resource text file and returns content as a list.

    Arguments:
        path: The resource path relative to the data root directory

    Returns:
        A list with the file contents.
    """
    lines = []
    with open(resource_path(path)) as source:
        lines = source.read().splitlines()

    return lines

def slicer(string, start, end):
    """ Slice a string

    Return a slice of the string matching anything between, and including,
    the `start` and `end` characters.

    Arguments:
        string: Original string
        start: Start character to begin slicing
        end: End character to stop slicing

    Returns:
        A new string which is a slice from the original string, including the start
        and end characters.

        None is return in case a valid slice couldn't be found, as in start or end
        characters are inversed or doesn't exist in the original slice.
    """
    result = ""
    add = False
    for char in string:
        if char == start:
            add = True
        if add:
            result += char

        # Only return if we have found start
        if char == end and add:
            return result

    return None

# pylint: disable=too-few-public-methods
class Generator:
    """ Base Generator class

    Subclasses of Generator must implement the generate() method.

    Attributes:
        rnd: Access to python built-in random module
        ascii_characters: String with valid ascii characters
        integers: String with valid integers
    """
    def __init__(self, rnd=None):
        self.rnd = rnd or random
        self.alpha = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
        self.digit = "0123456789"

    def generate(self):
        """ Generate data
        Must be implemented by subclasses
        """
        raise NotImplementedError()


# pylint: disable=too-few-public-methods
class Integer(Generator):
    """ Random Integer generator """

    def __init__(self, start=1, end=100, **kwargs):
        """ Initialize generator

        Arguments:
            start: Minimum value
            end: Maximum value

        """
        super().__init__(**kwargs)
        self.start = start
        self.end = end

    def generate(self):
        """ Generate a random integer

        Set the start/end attributes prior to calling this method.

        Returns:
            A single random integer from `start` to `end`.
        """
        return self.rnd.randint(self.start, self.end)


class Sequence(Generator):
    """ Sequence integer generator """

    def __init__(self, start=1, end=100, **kwargs):
        """ Initialize generator

        Arguments:
            start: Start of sequence
            end: End of sequence
        """
        super().__init__(**kwargs)
        self._start = start
        self.end = end

        self._next = self.start

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value):
        self._start = value
        if self._next < self._start:
            self._next = self._start

    def generate(self):
        """ Generate next integer in the sequence

        This method will raise a RuntimeError if the sequence
        has reached the end.
        """
        value = self._next
        self._next += 1

        if value > self.end:
            raise RuntimeError

        return value

# pylint: disable=too-few-public-methods
class Name(Generator):
    """ Random Name generator """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._lines = None

    def generate(self):
        """ Generate a random name

        Returns:
            A random name in string format
        """
        if self._lines is None:
            self._lines = read_resource("names/names.txt")

        result = self.rnd.choice(self._lines)

        return result

class String(Generator):
    """ Generate string from pattern

    This generator work by reading a pattern that describes how to
    generate the string.

    The pattern looks like a simplified regular expression but is processed
    completely different, so don't expect normal regular expressions to work.

    See docs for details.
    """

    def __init__(self, pattern=None, parser=None, **kwargs):
        super().__init__(**kwargs)
        self.pattern = pattern
        self.parser = parser or self._create_parser()

    def _create_parser(self): # pylint: disable=no-self-use
        tokenizer = Tokenizer()
        return SGParser(tokenizer=tokenizer)

    def _range(self, quantifier):
        """ Create a range from quantifier

        There are two types of quantifiers

        QUANTIFIER: Repeat X times
        QUANTIFIER_RANGE: Repeat X times, where X is anywhere between start and end

        Returns:
            Python built in range() with a size depending on quantifier type.
        """
        size = 1
        if quantifier["type"] == "QUANTIFIER":
            size = quantifier["value"]
        elif quantifier["type"] == "QUANTIFIER_RANGE":
            start = quantifier["value"]["start"]
            end = quantifier["value"]["end"]
            size = self.rnd.choice(list(range(start, end+1)))

        return range(size)

    def generate_CHARCODE(self, node): # pylint: disable=invalid-name
        """ Generate CHARCODE string

        Generates a string depending on the CHARCODE:
            "c": Character/alpha
            "d": Digit
        """
        result = ""

        for _ in self._range(node["repeat"]):
            if node["value"] == "c":
                result += self.rnd.choice(self.alpha)
            elif node["value"] == "d":
                result += self.rnd.choice(self.digit)
            else:
                raise ValueError("Invalid CHARCODE %s" % node["value"])

        return result

    def generate_ONEOF(self, node): # pylint: disable=invalid-name
        """ Generate one from list

        Returns a value from a list of valid values
        """
        result = ""

        for _ in self._range(node["repeat"]):
            result += self.rnd.choice(node["value"])

        return result

    def generate_RANGE(self, node): # pylint: disable=invalid-name
        """ Generate a range of values

        Generates, in sequence, a number of alpha or digit characters.
        """
        result = ""
        start = node["value"]["start"]
        end = node["value"]["end"]

        source = self.alpha
        if str.isdigit(start):
            source = self.digit

        for _ in self._range(node["repeat"]):
            choices = slicer(source, start, end)
            result += self.rnd.choice(choices)

        return result

    def generate_STRING_GROUP(self, node): # pylint: disable=invalid-name
        """ Generate a string form a list of strings """
        result = ""

        for _ in self._range(node["repeat"]):
            result += self.rnd.choice(node["value"])

        return result

    def generate_NUMBER(self, node): # pylint: disable=invalid-name
        """ Generate number literal """
        result = ""

        for _ in self._range(node["repeat"]):
            result += str(node["value"])

        return result

    def generate_STRING(self, node): # pylint: disable=invalid-name, no-self-use
        """ Generate string literal """
        return node["value"]

    def generate_LITERAL(self, node): # pylint: disable=invalid-name, no-self-use
        """ Generate literal """
        return node["value"]

    def generate(self):
        """ Generate a string based on pattern """
        ast = self.parser.parse(self.pattern)
        result = ""

        for node in ast:
            function_name = "generate_" + node["type"]
            if not hasattr(self, function_name):
                raise NotImplementedError("Unknown node type %s" % node["type"])

            func = getattr(self, function_name)
            result += func(node)

        return result



""" Module for seed cli subgroup

$ FLASK_APP=app flask seed <command>
"""

import os
import re
import importlib.util
import inspect
import click
from flask.cli import with_appcontext
from flask import current_app as app

from flask_seeder import Seeder

def get_seed_scripts(root="seeds"):
    """ Get seed scripts

    Recursively walk a directory structure and find all python scripts.

    Arguments:
        root: Optional root directory to start walking (Default: "seeds")

    Returns:
        Returns a list with python file paths to found python scripts.
        If no files are found, and empty list is returned.

    """
    extension = "py"
    result = []
    for path, _, files in os.walk(root):
        for filename in files:
            if not re.search(r"\."+extension+"$", filename):
                continue

            # Concat the full file path
            file_path = os.path.join(path, filename)

            result.append(file_path)

    return result

def inheritsfrom(child, parent):
    """ Verify inheritance

    Check if child inherits from parent. Does the same thing as built-in
    issubclass() except child and parent can't be the same class.

    Arguments:
        child: Child class
        parent: Parent class

    Returns:
        Returns True if Child inherits from Parent.
        Returns False if Child does not inherit from Parent or if Child and Parent
        is the same class.
    """
    return issubclass(child, parent) and child != parent

def load_seeder(cls, name, mod_path, file_path):
    """ Load seeder instance

    Instantiate a seeder using common base settings.

    Arguments:
        cls: Seeder class
        name: String name representation of the seeder
        mod_path: Python module path (package.module)
        file_path: Path to the python file in the file system

    Returns:
        Returns an instance of `cls`.
    """
    seeder = cls()
    seeder.name = name
    seeder.mod_path = mod_path
    seeder.file_path = file_path

    return seeder

def get_seeders_from_script(script):
    """ Get all seeders from a script file

    Reads a python script and detecs all classes within that inherits from the base Seeder class.

    Arguments:
        script: Filesystem path to the script

    Returns:
        Returns a list of loaded seeder objects from the script
    """
    seeders = []
    spec = importlib.util.spec_from_file_location("flask_seeder.seeder.ext", script)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    members = inspect.getmembers(module, inspect.isclass)
    for member in members:
        if inheritsfrom(member[1], Seeder):
            seeder = load_seeder(member[1], member[0], "", script)
            seeders.append(seeder)

    return seeders

def get_seeders(root=None):
    """ Get all seeders from all scripts

    Finds all python scripts with seeders, loads them and return them.

    Returns:
        List of loaded seeder objects
    """
    seeders = []
    if root is not None:
        scripts = get_seed_scripts(root=root)
    else:
        scripts = get_seed_scripts()

    for script in scripts:
        seeders.extend(get_seeders_from_script(script))

    return seeders


@click.group()
def seed():
    """ Database seed commands """

@seed.command("run")
@click.option("--root", default="seeds", type=click.Path(),
              help="Root directory for seed scripts",
              envvar="FLASK_SEEDER_ROOT")
@click.option("--commit/--no-commit", default=True,
              help="Commit changes to database after seeding",
              envvar="FLASK_SEEDER_AUTOCOMMIT")
@click.argument("seeders", nargs=-1)
@with_appcontext
def seed_run(root, commit, seeders):
    """ Run database seeders

    Any optional arguments after the options will be treated as a list of seeders to run,
    for example:

        $ flask seed run DemoSeeder AnotherDemoSeeder

    """
    click.echo("Running database seeders")
    db = None
    try:
        db = app.extensions["flask_seeder"].db
    except KeyError:
        raise RuntimeError("Flask-Seeder not initialized!")

    for seeder in get_seeders(root=root):
        if seeders and seeder.name not in seeders:
            continue

        seeder.db = db
        try:
            seeder.run()
        # pylint: disable=broad-except,invalid-name
        except Exception as e:
            click.echo("%s...\t[ERROR]" % seeder.name)
            click.echo("\t%s" % e)
            continue

        click.echo("%s...\t[OK]" % seeder.name)

    if not commit:
        click.echo("Not committing changes to database!")
        return

    click.echo("Committing to database!")
    db.session.commit()


@seed.command("list")
@click.option("--root", default="seeds", type=click.Path(),
              help="Root directory for seed scripts",
              envvar="FLASK_SEEDER_ROOT")
def seed_list(root):
    """ List all discoverable seeders """
    for seeder in get_seeders(root=root):
        click.echo("* %s" % seeder.name)



from unittest import TestCase

from flask_seeder.parser import Tokenizer

class TestTokenizer(TestCase):

    def setUp(self):
        self.t = Tokenizer()

    def test_tokenize_CHARCODE(self):
        self.t.run(r"\d")

        self.assertEqual(self.t.next().group, "CHARCODE")

    def test_tokenize_ONEOF(self):
        self.t.run("[abc]")

        self.assertEqual(self.t.next().group, "ONEOF")

    def test_tokenize_RANGE(self):
        self.t.run("[0-9]")

        self.assertEqual(self.t.next().group, "RANGE")

    def test_tokenize_QUANTIFIER(self):
        self.t.run("{1}")

        self.assertEqual(self.t.next().group, "QUANTIFIER")

    def test_tokenize_QUANTIFIER_RANGE(self):
        self.t.run("{1,5}")

        self.assertEqual(self.t.next().group, "QUANTIFIER_RANGE")

    def test_tokenize_STRING_GROUP(self):
        self.t.run("(one|two|three)")

        self.assertEqual(self.t.next().group, "STRING_GROUP")

    def test_tokenize_NUMBER(self):
        self.t.run("123")

        self.assertEqual(self.t.next().group, "NUMBER")

    def test_tokenize_STRING(self):
        self.t.run("string")

        self.assertEqual(self.t.next().group, "STRING")

    def test_tokenize_LITERAL(self):
        self.t.run("]#!.")

        self.assertEqual(self.t.next().group, "LITERAL")

    def test_tokenize_combination(self):
        expected = [
            "STRING",
            "RANGE",
            "QUANTIFIER",
            "CHARCODE",
            "STRING_GROUP",
            "QUANTIFIER_RANGE",
            "ONEOF",
            "QUANTIFIER",
            "NUMBER"
        ]

        self.t.run("abc[0-9]{5}\\d(one|two){1,2}[qwerty]{2}123")

        groups = [grp.group for grp in self.t.tokens]

        self.assertListEqual(groups, expected)

    def test_next_return_None_when_no_tokens_remain(self):
        self.assertIsNone(self.t.next())

    def test_peek_return_None_when_no_tokens_remain(self):
        self.assertIsNone(self.t.peek())

    def test_peek_return_next_token_without_moving_cursor(self):
        self.t.run("abc")

        self.assertEqual(self.t.peek(), self.t.peek())



import os
from unittest import TestCase
from unittest.mock import patch, MagicMock
from flask import Flask

from flask_seeder import cli, FlaskSeeder
from flask_seeder import Seeder

# This is for mocking os.walk, returns a list with dirs and files
# [ <current dir>, [<subdirs>], [<files in dir>] ]
MOCK_FILES = [
    ["data", ["sub1", "sub2"], []],
    [os.path.join("data", "sub1"), [], ["file1.py", "file2.py"]],
    [os.path.join("data", "sub2"), [], ["file3.py", "file4.py"]],
]

class TestSeedCLI(TestCase):

    def setUp(self):
        self.db_mock = MagicMock()

        self.app = Flask("test")
        self.cli = self.app.test_cli_runner()
        self.seeder = FlaskSeeder()
        self.seeder.init_app(self.app, db=self.db_mock)

    @patch("flask_seeder.cli.os.walk", return_value=MOCK_FILES)
    def test_get_seed_scripts_return_list_of_modules(self, mocked):
        expected_result = [
            os.path.join("data", "sub1", "file1.py"),
            os.path.join("data", "sub1", "file2.py"),
            os.path.join("data", "sub2", "file3.py"),
            os.path.join("data", "sub2", "file4.py"),
        ]

        modules = cli.get_seed_scripts()

        self.assertListEqual(modules, expected_result)

    @patch("flask_seeder.cli.get_seeders")
    def test_seeder_calls_run_method_on_loaded_modules(self, m_get_seeders):
        m_seeder = MagicMock()
        m_get_seeders.return_value = [m_seeder]

        self.cli.invoke(cli.seed_run)

        m_seeder.run.assert_called_once()

    @patch("flask_seeder.cli.get_seed_scripts", return_value=["test"])
    @patch("flask_seeder.cli.get_seeders_from_script")
    def test_get_seeders_return_list_of_seeders(self, m_get_seeders, m_get_scripts):
        m_seeder = MagicMock()
        m_get_seeders.return_value = [m_seeder]
        expected_result = [m_seeder]

        result = cli.get_seeders()

        self.assertListEqual(result, expected_result)

    @patch("flask_seeder.cli.get_seeders")
    def test_seed_list_print_list_of_seeders(self, m_get_seeders):
        m_seeder = MagicMock()
        m_seeder.name = "testseeder"
        m_get_seeders.return_value = [m_seeder]

        result = self.cli.invoke(cli.seed_list)

        self.assertTrue("testseeder" in result.output)

    @patch("flask_seeder.cli.get_seeders")
    def test_seed_run_echo_failed_runs(self, m_get_seeders):
        class TestSeeder(Seeder):
            def run(self):
                raise ValueError()
        m_get_seeders.return_value = [TestSeeder()]

        result = self.cli.invoke(cli.seed_run)

        self.assertTrue("ERROR" in result.output)

    @patch("flask_seeder.cli.get_seeders")
    def test_root_option_with_list(self, m_get_seeders):

        self.cli.invoke(cli.seed_list, args=["--root", "sub1"])

        m_get_seeders.assert_called_once_with(root="sub1")

    @patch("flask_seeder.cli.get_seeders")
    def test_root_option_with_run(self, m_get_seeders):

        self.cli.invoke(cli.seed_run, args=["--root", "sub1"])

        m_get_seeders.assert_called_once_with(root="sub1")

    @patch("flask_seeder.cli.get_seeders")
    def test_run_with_seeder_argument(self, m_get_seeders):
        m_seeder = MagicMock()
        m_seeder.name = "TestSeeder"
        m_get_seeders.return_value = [m_seeder]

        self.cli.invoke(cli.seed_run, args=["TestSeeder"])

        m_seeder.run.assert_called_once()

    @patch("flask_seeder.cli.get_seeders", return_value=[])
    def test_run_commit_session_by_default(self, m_get_seeders):
        self.cli.invoke(cli.seed_run)

        self.db_mock.session.commit.assert_called()

    @patch("flask_seeder.cli.get_seeders", return_value=[])
    def test_run_avoid_commit_with_no_commit_option(self, m_get_seeders):
        self.cli.invoke(cli.seed_run, args=["--no-commit"])

        self.assertFalse(self.db_mock.session.commit.called)



from unittest import TestCase
from unittest.mock import MagicMock

from flask_seeder.generator import Sequence

class TestSequenceGenerator(TestCase):

    def setUp(self):
        self.rnd_mock = MagicMock()
        self.generator = Sequence(rnd=self.rnd_mock)


    def test_generate_sequential_numbers(self):
        first = self.generator.generate()
        second = self.generator.generate()

        self.assertTrue(first+1 == second)

    def test_generate_sequential_start_end(self):
        self.generator.start = 23
        self.generator.end = 25

        first = self.generator.generate()
        second = self.generator.generate()

        self.assertEqual(first, 23)
        self.assertEqual(second, 24)

    def test_generate_raise_RuntimeError_when_exhausted(self):
        # Make start larger than end to trigger RuntimeError
        self.generator.start = 10
        self.generator.end = 5

        with self.assertRaises(RuntimeError):
            self.generator.generate()


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Flask-Seeder",
    version="1.0",
    url="https://github.com/diddi-/flask-seeder",
    author="Diddi Oskarsson",
    author_email="diddi@diddi.se",
    description="Flask extension to seed database through scripts",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    include_package_data=True,
    license="MIT",
    install_requires=[
        "Flask>=1.0.2",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "flask.commands": [
            "seed=flask_seeder.cli:seed",
        ]
    }
)



from unittest import TestCase

from flask_seeder import Faker
from flask_seeder.generator import Generator

class Dummy:
    def __init__(self, test_arg=None):
        self.test_arg = test_arg

class TestFaker(TestCase):

    def setUp(self):
        self.faker = Faker(cls=Dummy)

    def test_create_return_instance(self):
        result = self.faker.create()

        self.assertIsInstance(result[0], Dummy)

    def test_create_with_limit(self):
        result = self.faker.create(2)

        self.assertEqual(len(result), 2)

    def test_create_instantiate_with_init_args_value(self):
        self.faker.init = {"test_arg": "test_value"}

        result = self.faker.create()

        self.assertEqual(result[0].test_arg, "test_value")

    def test_create_run_init_args_generator(self):
        class DummyGenerator(Generator):
            def generate(self):
                return "test_value"
        self.faker.init = {"test_arg": DummyGenerator()}

        result = self.faker.create()

        self.assertEqual(result[0].test_arg, "test_value")



from unittest import TestCase
from flask_seeder import Seeder

class TestBaseSeeder(TestCase):

    def setUp(self):
        self.seeder = Seeder()

    def test_run_raise_NotImplementedError(self):
        with self.assertRaises(NotImplementedError):
            self.seeder.run()




from unittest import TestCase
from unittest.mock import MagicMock

from flask_seeder.parser import SGParser, Token, strtype


class TestUtility(TestCase):

    def test_strtype_return_DIGIT_type(self):
        result = strtype("123")

        self.assertEqual(result, "DIGIT")

    def test_strtype_return_ALPHA_type(self):
        result = strtype("abc")

        self.assertEqual(result, "ALPHA")

    def test_strtype_return_STRING_type(self):
        result = strtype("abc123#.}")

        self.assertEqual(result, "STRING")

class TestSGParser(TestCase):

    def setUp(self):
        self.tokenizer = MagicMock()
        self.parser = SGParser(tokenizer=self.tokenizer)

    def test_parse_string(self):
        expected = {"type": "STRING", "value": "qwerty"}
        token = Token("STRING", "qwerty")

        result = self.parser.parse_STRING(token)

        self.assertDictEqual(result, expected)

    def test_parse_range(self):
        expected = {"type": "RANGE", "value": {"start": "1", "end": "2"},
                    "repeat": {
                        "type": "QUANTIFIER",
                        "value": 1,
                    }}
        token = Token("RANGE", "[1-2]")

        result = self.parser.parse_RANGE(token)

        self.assertDictEqual(result, expected)

    def test_parse_range_raise_RuntimeError_for_inversed_range(self):
        token = Token("RANGE", "[2-1]")

        with self.assertRaises(ValueError):
            self.parser.parse_RANGE(token)

    def test_prase_range_raise_RuntimeError_for_mixed_start_end(self):
        token = Token("RANGE", "[1-z]")

        with self.assertRaises(ValueError):
            self.parser.parse_RANGE(token)


    def test_parse_quantifier(self):
        expected = {"type": "QUANTIFIER", "value": 5}
        token = Token("QUANTIFIER", "{5}")

        result = self.parser.parse_QUANTIFIER(token)

        self.assertDictEqual(result, expected)

    def test_parse_range_with_quantifier(self):
        expected = {"type": "RANGE",
                    "value": {
                        "start": "1",
                        "end": "2"
                    },
                    "repeat": {
                        "type": "QUANTIFIER",
                        "value": 5
                    }}
        range_token = Token("RANGE", "[1-2]")
        quant_token = Token("QUANTIFIER", "{5}")
        self.tokenizer.peek.return_value = quant_token
        self.tokenizer.next.return_value = quant_token

        result = self.parser.parse_RANGE(range_token)

        self.assertDictEqual(result, expected)

    def test_parse_range_with_quantifier_range(self):
        expected = {"type": "RANGE",
                    "value": {
                        "start": "1",
                        "end": "2"
                    },
                    "repeat": {
                        "type": "QUANTIFIER_RANGE",
                        "value": {
                            "start": 5,
                            "end": 10
                        }
                    }}
        range_token = Token("RANGE", "[1-2]")
        quant_token = Token("QUANTIFIER_RANGE", "{5,10}")
        self.tokenizer.peek.return_value = quant_token
        self.tokenizer.next.return_value = quant_token

        result = self.parser.parse_RANGE(range_token)

        self.assertDictEqual(result, expected)

    def test_parse_charcode(self):
        expected = {"type": "CHARCODE", "value": "d",
                    "repeat": {
                        "type": "QUANTIFIER",
                        "value": 1,
                    }}
        token = Token("CHARCODE", r"\d")

        result = self.parser.parse_CHARCODE(token)

        self.assertDictEqual(result, expected)

    def test_parse_charcode_with_quantifier(self):
        expected = {"type": "CHARCODE",
                    "value": "d",
                    "repeat": {
                        "type": "QUANTIFIER",
                        "value": 2}}
        range_token = Token("CHARCODE", r"\d")
        quant_token = Token("QUANTIFIER", "{2}")
        self.tokenizer.peek.return_value = quant_token
        self.tokenizer.next.return_value = quant_token

        result = self.parser.parse_CHARCODE(range_token)

        self.assertDictEqual(result, expected)

    def test_parse_charcode_with_quantifier_range(self):
        expected = {"type": "CHARCODE",
                    "value": "d",
                    "repeat": {
                        "type": "QUANTIFIER_RANGE",
                        "value": {
                            "start": 5,
                            "end": 10
                        }
                    }}
        range_token = Token("CHARCODE", r"\d")
        quant_token = Token("QUANTIFIER_RANGE", "{5,10}")
        self.tokenizer.peek.return_value = quant_token
        self.tokenizer.next.return_value = quant_token

        result = self.parser.parse_CHARCODE(range_token)

        self.assertDictEqual(result, expected)

    def test_parse_oneof(self):
        expected = {"type": "ONEOF", "value": ["a", "b", "c"],
                    "repeat": {
                        "type": "QUANTIFIER",
                        "value": 1,
                    }}
        token = Token("ONEOF", "[abc]")

        result = self.parser.parse_ONEOF(token)

        self.assertDictEqual(result, expected)

    def test_parse_oneof_with_quantifier(self):
        expected = {"type": "ONEOF",
                    "value": ["a", "b", "c"],
                    "repeat": {
                        "type": "QUANTIFIER",
                        "value": 2
                    }}
        range_token = Token("ONEOF", "[abc]")
        quant_token = Token("QUANTIFIER", "{2}")
        self.tokenizer.peek.return_value = quant_token
        self.tokenizer.next.return_value = quant_token

        result = self.parser.parse_ONEOF(range_token)

        self.assertDictEqual(result, expected)

    def test_parse_oneof_with_quantifier_range(self):
        expected = {"type": "ONEOF",
                    "value": ["a", "b", "c"],
                    "repeat": {
                        "type": "QUANTIFIER_RANGE",
                        "value": {
                            "start": 5,
                            "end": 10
                        }
                    }}
        range_token = Token("ONEOF", "[abc]")
        quant_token = Token("QUANTIFIER_RANGE", "{5,10}")
        self.tokenizer.peek.return_value = quant_token
        self.tokenizer.next.return_value = quant_token

        result = self.parser.parse_ONEOF(range_token)

        self.assertDictEqual(result, expected)

    def test_parse_quantifier_range(self):
        expected = {"type": "QUANTIFIER_RANGE", "value": {"start": 1, "end": 2}}
        token = Token("QUANTIFIER_RANGE", "{1,2}")

        result = self.parser.parse_QUANTIFIER_RANGE(token)

        self.assertDictEqual(result, expected)

    def test_parse_quantifier_range_raise_RuntimeError_for_inversed_range(self):
        token = Token("QUANTIFIER_RANGE", "{2,1}")

        with self.assertRaises(ValueError):
            self.parser.parse_QUANTIFIER_RANGE(token)

    def test_parse_string_group(self):
        expected = {"type": "STRING_GROUP", "value": ["one", "two", "three"],
                    "repeat": {
                        "type": "QUANTIFIER",
                        "value": 1,
                    }}
        token = Token("STRING_GROUP", "(one|two|three)")

        result = self.parser.parse_STRING_GROUP(token)

        self.assertDictEqual(result, expected)

    def test_parse_string_group_with_quantifier(self):
        expected = {"type": "STRING_GROUP",
                    "value": ["one", "two"],
                    "repeat": {
                        "type": "QUANTIFIER",
                        "value": 3
                    }}
        range_token = Token("STRING_GROUP", "(one|two)")
        quant_token = Token("QUANTIFIER", "{3}")
        self.tokenizer.peek.return_value = quant_token
        self.tokenizer.next.return_value = quant_token

        result = self.parser.parse_STRING_GROUP(range_token)

        self.assertDictEqual(result, expected)

    def test_parse_string_group_with_quantifier_range(self):
        expected = {"type": "STRING_GROUP",
                    "value": ["one", "two"],
                    "repeat": {
                        "type": "QUANTIFIER_RANGE",
                        "value": {
                            "start": 5,
                            "end": 10
                        }
                    }}
        range_token = Token("STRING_GROUP", "(one|two)")
        quant_token = Token("QUANTIFIER_RANGE", "{5,10}")
        self.tokenizer.peek.return_value = quant_token
        self.tokenizer.next.return_value = quant_token

        result = self.parser.parse_STRING_GROUP(range_token)

        self.assertDictEqual(result, expected)

    def test_parse_number(self):
        expected = {"type": "NUMBER", "value": 123,
                    "repeat": {
                        "type": "QUANTIFIER",
                        "value": 1,
                    }}
        token = Token("NUMBER", "123")

        result = self.parser.parse_NUMBER(token)

        self.assertDictEqual(result, expected)

    def test_parse_number_with_quantifier(self):
        expected = {"type": "NUMBER",
                    "value": 123,
                    "repeat": {
                        "type": "QUANTIFIER",
                        "value": 3
                    }}
        range_token = Token("NUMBER", "123")
        quant_token = Token("QUANTIFIER", "{3}")
        self.tokenizer.peek.return_value = quant_token
        self.tokenizer.next.return_value = quant_token

        result = self.parser.parse_NUMBER(range_token)

        self.assertDictEqual(result, expected)

    def test_parse_number_with_quantifier_range(self):
        expected = {"type": "NUMBER",
                    "value": 123,
                    "repeat": {
                        "type": "QUANTIFIER_RANGE",
                        "value": {
                            "start": 5,
                            "end": 10
                        }
                    }}
        range_token = Token("NUMBER", "123")
        quant_token = Token("QUANTIFIER_RANGE", "{5,10}")
        self.tokenizer.peek.return_value = quant_token
        self.tokenizer.next.return_value = quant_token

        result = self.parser.parse_NUMBER(range_token)

        self.assertDictEqual(result, expected)

    def test_parse_literal(self):
        expected = {"type": "LITERAL", "value": "#(]"}
        token = Token("LITERAL", "#(]")

        result = self.parser.parse_LITERAL(token)

        self.assertDictEqual(result, expected)
